// Program takes an existing pixel file and bins image using it
// Run with --help to get command line options

//      Described in Sanders and Fabian (submitted)
//      Routines for adaptively binning data
//      Copyright (C) 2000 Jeremy Sanders
//      Contact: jss@ast.cam.ac.uk
//               Institute of Astronomy, Madingley Road,
//               Cambridge, CB3 0HA, UK.

//      See the file COPYING for full licence details.

//      This program is free software; you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation; either version 2 of the License, or
//      (at your option) any later version.

//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.

//      You should have received a copy of the GNU General Public License
//      along with this program; if not, write to the Free Software
//      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include <iostream>
#include <string>
#include <strstream>
#include <cstdlib>
#include <cmath>
#include <FITSFile.h>
#include <FITSImage.h>
#include <FITSPosn.h>
#include <parammm/parammm.hh>
#include "binmodule.hh"

class ABPixelCopyProgram
{
public:
  ABPixelCopyProgram(int argc, char *argv[]);
  ~ABPixelCopyProgram();

  void Run();

private:
  void binimage();

private:
  string m_infname, m_pixelfname, m_outfname, m_errfname;

  CFITSImage m_pixelImage;
  CFITSImage m_inImage;
  CFITSImage m_outImage, m_errImage;
  CFITSPosn m_inPosn;
  double m_backgrnd;
};

ABPixelCopyProgram::ABPixelCopyProgram(int argc, char *argv[])
  : m_pixelfname("pixel.fits"),
    m_backgrnd(0.)
{
  parammm::param params(argc, argv);
  params.add_switch( parammm::pswitch("image", 'i',
				      parammm::pstring_opt(&m_infname),
				      "set input image file (req)",
				      "FILE"));
  params.add_switch( parammm::pswitch("out", 'o',
				      parammm::pstring_opt(&m_outfname),
				      "set out file (req)",
				      "FILE"));
  params.add_switch( parammm::pswitch("err", 'e',
				      parammm::pstring_opt(&m_errfname),
				      "set out error file (req)",
				      "FILE"));
  params.add_switch( parammm::pswitch("pixel", 'p',
				      parammm::pstring_opt(&m_pixelfname),
				      "set input pixel file (def pixel.fits)",
				      "FILE"));
  params.add_switch( parammm::pswitch("background", 'b',
				      parammm::pdouble_opt(&m_backgrnd),
				      "set background counts/pixel",
				      "VAL"));
  params.set_autohelp("Usage: ABPixelCopy [OPTIONS] --image=in.fits "
		      "--out=out.fits --err=err.fits\n"
		      "Bins image using pixel file, creating output and error "
		      "file.\nDoes background subtraction.\n"
		      "Written by Jeremy Sanders, 2000.",
		      "Report bugs to <jss@ast.cam.ac.uk>");
  params.enable_autohelp();
  params.interpret_and_catch();

  if( m_infname.empty() || m_outfname.empty() || m_errfname.empty() )
    params.show_autohelp();
}

ABPixelCopyProgram::~ABPixelCopyProgram()
{
}

void ABPixelCopyProgram::Run()
{
  // load images
  {
    CFITSFile infile(m_infname.c_str(), CFITSFile::existingro);
    m_inImage = infile.GetImage();
    m_inPosn = infile.GetPosn();
  }{
    CFITSFile infile(m_pixelfname.c_str(), CFITSFile::existingro);
    m_pixelImage = infile.GetImage();
  }

  binimage();

  // save image
  {
    CFITSFile outfile(m_outfname.c_str(), CFITSFile::create);
    outfile.SetImage(m_outImage);
    outfile.SetPosn(m_inPosn);
    outfile.WriteImage();
  }{
    CFITSFile outfile(m_errfname.c_str(), CFITSFile::create);
    outfile.SetImage(m_errImage);
    outfile.SetPosn(m_inPosn);
    outfile.WriteImage();
  }
}

void ABPixelCopyProgram::binimage()
{
  // rewritten to avoid going through the image too many times

  const int xw = m_inImage.GetXW();
  const int yw = m_inImage.GetYW();

  m_outImage.Resize(xw, yw);
  m_errImage.Resize(xw, yw);

  int maxpix = -1;
  for(int y=0; y<yw; y++)
    for(int x=0; x<xw; x++) {
      const int pix = (int)m_pixelImage.GetPixel(x, y);
      if( pix > maxpix ) maxpix = pix;
    }
  vector<AdaptiveBin::pixlist> pixels(maxpix+1);

  for(int y=0; y<yw; y++)
    for(int x=0; x<xw; x++) {
      const int pix = (int)m_pixelImage.GetPixel(x, y);
      pixels[pix].push_back( AdaptiveBin::pixel(x, y) );
    }

  // do the binning
  for(int pixelno=maxpix; pixelno>=0; --pixelno) {
    double tot = 0.;
    AdaptiveBin::pixlist &p = pixels[pixelno];

    if( p.size() == 0 ) {
      cerr << "Aarrgghh - missing pixel in pixel image\n";
      exit(1);
    }

    for(int i=p.size()-1; i>=0; --i) {
      const double pix = m_inImage.GetPixel( p[i].x(), p[i].y() );
      tot += pix;
    }

    const double av = tot/p.size() - m_backgrnd;
    const double ferr = sqrt(tot + m_backgrnd*p.size()) /
      (tot - m_backgrnd*p.size());

    for(int i=p.size()-1; i>=0; --i) {
      m_outImage.SetPixel(p[i].x(), p[i].y(), av);
      m_errImage.SetPixel(p[i].x(), p[i].y(), ferr);
    }
  } // for

}

int main(int argc, char *argv[])
{
  ABPixelCopyProgram program(argc, argv);
  program.Run();
  return 0;
}
