//      Adaptive Binning Program
//      Main program module
//      Described in Sanders and Fabian (submitted)
//      Routines for adaptively binning data
//      Copyright (C) 2000 Jeremy Sanders
//      Contact: jss@ast.cam.ac.uk
//               Institute of Astronomy, Madingley Road,
//               Cambridge, CB3 0HA, UK.

//      See the file COPYING for full licence details.

//      This program is free software; you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation; either version 2 of the License, or
//      (at your option) any later version.

//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.

//      You should have received a copy of the GNU General Public License
//      along with this program; if not, write to the Free Software
//      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include <iostream>
#include <algorithm>
#include <string>
#include <parammm/parammm.hh>
#include <FITSFile.h>

#include "binmodule.hh"

using std::string;
using std::sort;

namespace AdaptiveBin {

  class pixelval {
  public:
    pixelval(const pixlist &pl, double val);
    bool operator< (const pixelval &cmp) const;
    const pixlist& pixels() const { return m_pl; }
    double val() const { return m_val; }
  private:
    pixlist m_pl;
    double m_val;
  };

  pixelval::pixelval(const pixlist &pl, double val)
  {
    m_pl = pl;
    m_val = val;
  }

  bool pixelval::operator<(const pixelval &cmp) const
  {
    return m_val < cmp.m_val;
  }

  class binner
  {
  public:
    binner(binmodule *bm, double threshold,
	   int subpixposn);
    ~binner();

    void bin(CFITSImage *outimage,
	     CFITSImage *errorimage,
	     CFITSImage *piximage);

  private:
    //    void binpass(int pixsize, bool finalpass);
    void binpasssort(int pixsize, bool finalpass);
    
  private:
    int m_pixelno;

    double m_threshold;
    int m_subpixposn;
    binmodule *m_binmod;
    CFITSImage m_outimage;
    CFITSImage m_errorimage;
    CFITSImage m_handledimage;
  };

  binner::binner(binmodule *bm, double threshold, int subpixposn)
    : m_threshold(threshold),
      m_subpixposn(subpixposn),
      m_binmod(bm),
      m_outimage(bm->xw(), bm->yw()),
      m_errorimage(bm->xw(), bm->yw()),
      m_handledimage(bm->xw(), bm->yw())
  {
  }

  binner::~binner()
  {
  }

// old code - actually faster for non-overlapping pixels
//    void binner::binpass(int size, bool finalpass)
//    {
//      cout << "Pass " << size << endl;
//      // subpixel positioning for pixels code
//      int nx, ny, ns;
//      if( size < m_subpixposn ) {
//        nx = m_binmod->xw() + 1;
//        ny = m_binmod->yw() + 1;
//        ns = 1;
//      } else {
//        ns = size / m_subpixposn;
//        nx = m_binmod->xw() / ns + 1;
//        ny = m_binmod->yw() / ns + 1;
//      }

//      // iterate over subpixels
//      for(int x=0; x<nx; x++)
//        for(int y=0; y<ny; y++) {

//  	// make 'subpixel' with size size x size
//  	vector<pixel> pixels;
//  	for(int sx=0; sx < size; sx++)
//  	  for(int sy=0; sy < size; sy++) {
//  	    const int tx = x*ns+sx;
//  	    const int ty = y*ns+sy;
//  	    if(tx < m_binmod->xw() && ty < m_binmod->yw() )
//  	      if( m_handledimage.GetPixel(tx, ty) < 0. )
//  		pixels.push_back( pixel(tx, ty) );
//  	  }
//  	// end 'sub pixel'

//  	// are there any pixels in sub-pixel?
//  	if( pixels.size() > 0 ) {
//  	  // is binning error < threshold
//  	  const double error = m_binmod -> fracerror(pixels, true);
//  	  if(error <= m_threshold || finalpass) {

//  	    // set output pixels to correct thing
//  	    const double val = m_binmod -> value(pixels);
//  	    const double outerror = m_binmod -> fracerror(pixels, false);
//  	    for(int i=pixels.size()-1; i>=0; i--) {
//  	      const int x = pixels[i].x(), y = pixels[i].y();
//  	      m_handledimage.SetPixel(x, y, m_pixelno);
//  	      m_outimage.SetPixel(x, y, val);
//  	      m_errorimage.SetPixel(x, y, outerror);
//  	    }
//  	    m_pixelno ++;
//  	  }
//  	}

//        } // pixels
//      cout << endl;
//    } // fn

  void binner::bin(CFITSImage *outimage,
		   CFITSImage *errorimage,
		   CFITSImage *pixelimage)
  {
    m_handledimage.SetAll(-1.);
    m_pixelno = 0;

    int pass;
    for(pass=1; pass<m_binmod->xw() || pass<m_binmod->yw(); pass *= 2)
      binpasssort(pass, false);

    binpasssort(pass, true);
    *outimage = m_outimage;
    *errorimage = m_errorimage;
    *pixelimage = m_handledimage;
  }

  // sorting binning version of binpass
  void binner::binpasssort(int size, bool finalpass)
  {
    cout << "Pass " << size << endl;
    // code to allow pixels to start on sub-pixel boundries
    int nx, ny, ns;
    if( size < m_subpixposn ) {
      nx = m_binmod->xw() + 1;
      ny = m_binmod->yw() + 1;
      ns = 1;
    } else {
      ns = size / m_subpixposn;
      nx = m_binmod->xw() / ns + 1;
      ny = m_binmod->yw() / ns + 1;
    }

    // make an array of pixels with error less than threshold
    vector<pixelval> pixelslist;
    // iterate over subpixels
    for(int x=0; x<nx; x++)
      for(int y=0; y<ny; y++) {
	  
	// make 'subpixel' with size size x size
	pixlist pixels;
	for(int sx=0; sx < size; sx++)
	  for(int sy=0; sy < size; sy++) {
	    const int tx = x*ns+sx;
	    const int ty = y*ns+sy;
	    if(tx < m_binmod->xw() && ty < m_binmod->yw() )
	      if( m_handledimage.GetPixel(tx, ty) < 0. )
		pixels.push_back( pixel(tx, ty) );
	  }
	// end 'sub pixel'
	
	// are there any pixels in sub-pixel?
	if( pixels.size() > 0 ) {
	  // is binning error < threshold
	  const double error = m_binmod -> fracerror(pixels, true);
	  if(error <= m_threshold || finalpass) {
	    pixelval p(pixels, error);
	    pixelslist.push_back(p);
	  }
	}	      
	
      } // pixels

    // sort pixels into error order
    // (not needed unless subpixelling on)
    if(m_subpixposn != 1)
      sort(pixelslist.begin(), pixelslist.end());

    // set pixels which haven't already been set
    // with the values
    const int mi=pixelslist.size();
    for(int i=0; i<mi; i++) {
      // set output pixels to correct thing
      const pixlist &minerrpixel = pixelslist[i].pixels();

      bool spoilt = false;
      for(int j=minerrpixel.size()-1; j>=0 && !spoilt; j--) {
	if( ! (m_handledimage.GetPixel(minerrpixel[j].x(),
				       minerrpixel[j].y() ) < 0.) )
	  spoilt = true;
      }
      if(spoilt) continue;

      const double val = m_binmod -> value(minerrpixel);
      const double outerror = m_binmod -> fracerror(minerrpixel, false);
      for(int i=minerrpixel.size()-1; i>=0; i--) {
	const int x = minerrpixel[i].x(), y = minerrpixel[i].y();
	m_handledimage.SetPixel(x, y, m_pixelno);
	m_outimage.SetPixel(x, y, val);
	m_errorimage.SetPixel(x, y, outerror);
      }
      m_pixelno ++;
    }

  } // fn

} // namespace


///////////////////////////////////////////////////////////////////////
// Program class

class prog
{
public:
  prog(int argc, char **argv);
  ~prog();
  void run();

public:
  AdaptiveBin::binmodule *m_binmod;
  double m_threshold;
  string m_outfname, m_errfname, m_pixfname, m_value;
  int m_subpix;
};

prog::prog(int argc, char **argv)
  : m_binmod(0),
    m_threshold(0.1),
    m_outfname("adbin-out.fits"),
    m_errfname("adbin-err.fits"),
    m_pixfname("pixel.fits"),
    m_value("count(0)"),
    m_subpix(1)
{
  parammm::param params(argc, argv);
  params.add_switch( parammm::pswitch("out", 'o',
				      parammm::pstring_opt(&m_outfname),
				      "set out file",
				      "FILE"));
  params.add_switch( parammm::pswitch("error", 'e',
				      parammm::pstring_opt(&m_errfname),
				      "set error out file",
				      "FILE"));
  params.add_switch( parammm::pswitch("pixel", 'p',
				      parammm::pstring_opt(&m_pixfname),
				      "set pixel out file",
				      "FILE"));
  params.add_switch( parammm::pswitch("threshold", 't',
				      parammm::pdouble_opt(&m_threshold),
				      "set threshold fraction (def 10%)",
				      "VAL"));
  params.add_switch( parammm::pswitch("value", 'v',
				      parammm::pstring_opt(&m_value),
				      "set output value (eg count(0), "
				      "ratio(1,2))", "STR") );
  params.add_switch( parammm::pswitch("subpix", 's',
				      parammm::pint_opt(&m_subpix),
				      "set subpixel positioning "
				      "divisior (def. 1)",
				      "INT"));

  params.set_autohelp("Usage: AdaptiveBin [OPTIONS] file bg=count...\n"
		      "Adaptively bins a set of images\n"
		      "Written by Jeremy Sanders, 2000.",
		      "Report bugs to <jss@ast.cam.ac.uk>");
  params.enable_autohelp();
  params.interpret_and_catch();

  if(params.args().size() < 1)
    params.show_autohelp();

  try {
    m_binmod = new AdaptiveBin::ratio_binmodule(params.args());
    m_binmod->selectvalue(m_value);
  }
  catch(AdaptiveBin::invalidargs_exception e) {
    clog << "Invalid files listed\n\n";
    params.show_autohelp();
  }
  catch(AdaptiveBin::invalidvalue_exception e) {
    clog << "Invalid output value\n\n";
    params.show_autohelp();
  }

  cout << "Using value ";
  m_binmod->showvalue();
  cout << endl;
}

prog::~prog()
{
  if(m_binmod != 0)
    delete m_binmod;
}

void prog::run()
{
  CFITSImage out, err, pixel;
  AdaptiveBin::binner b(m_binmod, m_threshold, m_subpix);
  b.bin(&out, &err, &pixel);

  CFITSPosn posn;
  m_binmod -> getposn(&posn);

  {
    CFITSFile outf(m_outfname.c_str(), CFITSFile::create);
    outf.SetImage(out);
    outf.SetPosn(posn);
    outf.WriteImage();
  }{
    CFITSFile outf(m_errfname.c_str(), CFITSFile::create);
    outf.SetImage(err);
    outf.SetPosn(posn);
    outf.WriteImage();
  }{
    CFITSFile outf(m_pixfname.c_str(), CFITSFile::create);
    outf.SetImage(pixel);
    outf.SetPosn(posn);
    outf.WriteImage();
  }

}

int main(int argc, char *argv[])
{
  prog program(argc, argv);
  program.run();
  return 0;
}
